@@ -3,14 +3,23 @@ from django.template.loader import render_to_string
 from django.template import RequestContext, Context
 
 import hashlib
+import os
 
 class NexusModule(object):
+    # base url (pattern name) to show in navigation
     home_url = None
+    # generic permission required
+    permission = None
+    media_root = None
     
-    def __init__(self, site, name=None, app_name=None):
+    def __init__(self, site, category=None, name=None, app_name=None):
+        self.category = category
         self.site = site
         self.name = name
         self.app_name = app_name
+        if not self.media_root:
+            mod = __import__(self.__class__.__module__)
+            self.media_root = os.path.normpath(os.path.join(os.path.dirname(mod.__file__), 'media'))
 
     def render_to_string(self, template, context={}, request=None):
         if request:
